# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

import habana_frameworks.torch.core as htcore
import numpy as np
import torch

from .._quant_common.quant_config import MeasureExclude, QuantMode, ScaleMethod, get_hqt_config, set_hqt_config
from ..utils.logger import logger
from .common import *

imod_dict = {}
gmod_list = []


def patch_module_measure(mod, mconfig, mod_dict):
    parent = parent_child_mod_dict[mod].parent
    name = parent_child_mod_dict[mod].name
    patched_mod = mod_dict[mod.__class__.__name__].patched_module(mod, mconfig, name)
    setattr(parent, name, patched_mod)
    return patched_mod


def init_measure_object(mod, name, observer_class, mod_type, skip_measure_output, d_shape=None, params=None):
    input_observer = [
        observer_class(
            "input%d" % (i),
            mod,
            None if d_shape is None else d_shape.inputs[i],
            params=None if params is None else params.inputs[i],
        )
        for i in range(mod_type.num_inputs)
    ]
    if (not mod_type.required_output) and skip_measure_output:
        # excluding output measurements
        output_observer = None
    else:
        output_observer = [
            observer_class(
                "output%d" % (i),
                mod,
                None if d_shape is None else d_shape.outputs,
                params=None if params is None else params.outputs,
            )
            for i in range(mod_type.num_outputs)
        ]
    params_observer = {
        param_name: observer_class(
            param_name,
            mod,
            None if d_shape is None else d_shape.params[param_name],
            params=None if params is None else params.params[param_name],
        )
        for param_name in mod_type.param_names
    }
    measure_object = ModuleExtraConfig(input_observer, output_observer, params_observer, None, params)
    return measure_object


def prepare_model(model, mod_list=None):
    config = get_hqt_config(model).cfg
    observer_class = observer_types[config["observer"]]
    if (config["shape_file"] is not None) and (observer_class != ShapeObserver):
        shapes_fname = config["shape_file"] + ".json"
        d_shapes = load_file(shapes_fname, ShapeList, False)
    else:
        d_shapes = None
    gmod_list.extend(mod_list)
    generate_model_info(model)
    register_patched_measure_modules(model, mod_list, observer_class, d_shapes)


def register_patched_measure_modules(model, mod_list, observer_class, d_shapes=None):
    top_level_config = get_hqt_config(model)
    config = top_level_config.cfg
    skip_outputs_measurements = config["measure_exclude"] & (MeasureExclude.OUTPUT | MeasureExclude.ALL)
    patched_types = set()
    non_patched_types = set()
    patched_modules = []
    with torch.no_grad():
        for name, mod in model.named_modules():
            if (name in mod_list) or (mod_list is None):
                imod_dict[mod] = name
                mod_type_str = mod.__class__.__name__
                mod_type = config["mod_dict"][mod_type_str]
                params = (
                    observer_params[config["observer"]][mod_type]
                    if (config["observer"] in observer_params) and (mod_type in observer_params[config["observer"]])
                    else None
                )
                patched_types.add(type(mod))

                mod_extra_config = init_measure_object(
                    mod,
                    name,
                    observer_class,
                    mod_types[mod_type],
                    skip_outputs_measurements,
                    (d_shapes[name] if ((d_shapes is not None) and (name in d_shapes)) else None),
                    params,
                )
                set_hqt_config(mod, top_level_config)  # set config in the module, as it consumed by the patched module
                pmod = patch_module_measure(mod, mod_extra_config, mod_default_dict)
                for param_name in pmod._mod_extra_config.params:
                    param = getattr(pmod, param_name)
                    param = param.to("hpu")
                    pmod._mod_extra_config.params[param_name].measure(param)
                    htcore.mark_step()
                if observer_class == SaveObserver:
                    save_module(pmod)
                patched_modules.append(name)
            else:
                non_patched_types.add(type(mod))
    logger.debug("Patched module types: %s", patched_types)
    logger.debug("None-patched module types: %s", non_patched_types)
    logger.debug("Patched modules: %s", patched_modules)
    logger.debug("Total patched modules: %d", len(patched_modules))
    model = model.to("hpu")
    htcore.mark_step()


def is_measure_done(mod_extra_config):
    # check if measurements were collected by observer
    for obs in ([] if mod_extra_config.inputs is None else mod_extra_config.inputs) + (
        [] if mod_extra_config.outputs is None else mod_extra_config.outputs
    ):
        if obs.is_used():
            return True
    return False


def get_mod_extra_config_dict(model):
    mcd = {}
    for name, mod in model.named_modules():
        if hasattr(mod, "_mod_extra_config"):
            if is_measure_done(mod._mod_extra_config):
                name = name.replace("_orig_mod.", "")  # remove _orig_mod part added by dynamo mechanism
                mcd[name] = mod._mod_extra_config
            else:
                logger.debug(
                    "Layer '%s' has no measurements therefore it can't be quantized during quantization.",
                    name,
                )
    return mcd


def measure_control_to_state_dict(mcd):
    sd = {}
    sdl = {}
    for mname in mcd:
        sd[mname] = dict()
        sdl[mname] = dict()
        sd[mname]["inputs"] = [
            mcd[mname].inputs[i].state.detach().cpu().float().numpy()
            for i in range(len(mcd[mname].inputs))
            if mcd[mname].inputs[i].state is not None
        ]
        sdl[mname]["inputs"] = [
            mcd[mname].inputs[i].state.detach().cpu().float().numpy().tolist()
            for i in range(len(mcd[mname].inputs))
            if mcd[mname].inputs[i].state is not None
        ]
        if mcd[mname].outputs:
            sd[mname]["outputs"] = [
                mcd[mname].outputs[i].state.detach().cpu().float().numpy()
                for i in range(len(mcd[mname].outputs))
                if mcd[mname].outputs[i].state is not None
            ]
            sdl[mname]["outputs"] = [
                mcd[mname].outputs[i].state.detach().cpu().float().numpy().tolist()
                for i in range(len(mcd[mname].outputs))
                if mcd[mname].outputs[i].state is not None
            ]
        if len(mcd[mname].params) > 0:
            sd[mname]["params"] = dict()
            sdl[mname]["params"] = dict()
            for param_name in mcd[mname].params:
                if mcd[mname].params[param_name].state is not None:
                    sd[mname]["params"][param_name] = mcd[mname].params[param_name].state.detach().cpu().float().numpy()
                    sdl[mname]["params"][param_name] = (
                        mcd[mname].params[param_name].state.detach().cpu().float().numpy().tolist()
                    )
    return sd, sdl


def save_measurements(model, fname=None):
    config = get_hqt_config(model).cfg
    if config["mode"] in [QuantMode.MEASURE, QuantMode.SHAPE]:
        if fname is None:
            if ("measure_file" in config) and (config["measure_file"] is not None):
                fname_base = config["measure_file"]
                measure_type = "DynamicRange"
            elif ("shape_file" in config) and (config["shape_file"] is not None) and (config["observer"] == "shape"):
                fname_base = config["shape_file"]
                measure_type = "Shape"
            fname_np = fname_base + ".npz"
            fname_list = fname_base + ".json"
        else:
            logger.warning("'fname' is not None - Measurements/Shapes will not be saved")
            return
        mcd = get_mod_extra_config_dict(model)
        sd, sdl = measure_control_to_state_dict(mcd)

        logger.info("Dumping measurements")
        save_file(model, sd, np.ndarray, fname_np, measure_type)
        save_file(model, sdl, list, fname_list, measure_type)
        save_json(gmod_list, fname_base + "_mod_list.json")


def load_measurements(model, fname):
    config = get_hqt_config(model).cfg
    source_fname = fname if fname is not None else config["measure_file"]
    fname_np = source_fname + ".npz"
    d = load_file(
        fname_np,
        np.ndarray,
        fail_on_file_not_exist=(config["scale_method"] != ScaleMethod.UNIT_SCALE),
    )
    from collections import defaultdict

    d = defaultdict(lambda: None, d)

    return d


def save_json(d, fname):
    with open(fname, "w") as f:
        json.dump(d, f, indent=4)


def load_json(fname):
    with open(fname, "r") as f:
        d = json.load(f)
    return d


class MaxAbsObserver:
    def __init__(self, name, mod, d_shape=None, params=None):
        self.name = name
        self.mod = mod
        self.first = True
        self.used = False
        self.state = self.init_state_from_shape(d_shape)

    def init_state(self, x):
        device = x.device
        state = torch.zeros((1, 1), device=device, dtype=torch.float32)
        self.shape = list(x.shape)
        return state

    def init_state_from_shape(self, x_shape, device="hpu"):
        state = torch.zeros((1, 1), device=device, dtype=torch.float32)
        self.first = False
        return state

    def update_state(self, x):
        self.state.copy_(torch.maximum(torch.max(torch.abs(x)), self.state))

    def measure(self, x):
        if self.first:
            self.state = self.init_state(x)
            self.first = False
        self.update_state(x)
        self.used = True

    def is_used(self):
        return self.used


class MaxAbsPerChannelObserver:
    def __init__(self, name, mod, d_shape=None, params=None):
        self.name = name
        self.mod = mod
        self.first = True
        self.state = None
        self.used = False
        self.dim = params["dim"] if (params is not None) and ("dim" in params) else -1
        if d_shape is not None:
            p = list(range(len(d_shape)))
            self.dim = self.dim if self.dim >= 0 else len(d_shape) + self.dim
            p[-1] = self.dim
            p[self.dim] = len(d_shape) - 1
            self.p = p
            self.state = self.init_state_from_shape(d_shape)

    def init_state(self, x):
        device = x.device
        Nch = x.shape[self.dim]
        self.Nch = Nch
        state = torch.zeros((Nch, 1), device=device, dtype=torch.float32)
        self.shape = list(x.shape)
        return state

    def init_state_from_shape(self, x_shape, device="hpu"):
        device = device
        Nch = x_shape[self.dim]
        self.Nch = Nch
        state = torch.zeros((Nch, 1), device=device, dtype=torch.float32)
        self.first = False
        return state

    def update_state(self, x):
        self.state.copy_(
            torch.maximum(
                torch.max(
                    torch.abs(x.permute(self.p).reshape([-1, self.Nch])),
                    dim=0,
                    keepdim=True,
                )[0].t(),
                self.state,
            )
        )

    def measure(self, x):
        if self.first:
            self.state = self.init_state(x)
            self.first = False
        self.update_state(x)
        self.used = True

    def is_used(self):
        return self.used


def save_module(mod):
    folder_name = os.path.join(mod.config["dump_stats_base_path"], "tensors")
    os.makedirs(folder_name, exist_ok=True)
    file_base_name = os.path.join(folder_name, imod_dict[mod] + "_module.pt")
    torch.save(mod.state_dict(), file_base_name)


class SaveObserver:
    def __init__(self, name, mod, d_shape=None, params=None):
        self.name = name
        self.mod = mod
        self.first = True
        self.cnt = -1
        self.folder_name = os.path.join(config["dump_stats_base_path"], "tensors")
        os.makedirs(self.folder_name, exist_ok=True)
        self.file_base_name = os.path.join(self.folder_name, imod_dict[mod] + "_" + name + "_iter")
        self.state = self.init_state_from_shape(d_shape)
        self.used = False

    def init_state(self, x):
        device = x.device
        state = torch.zeros((1, 1), device=device, dtype=torch.float32)
        self.shape = list(x.shape)
        return state

    def init_state_from_shape(self, x_shape, device="hpu"):
        state = torch.zeros((1, 1), device=device, dtype=torch.float32)
        self.first = False
        return state

    def update_state(self, x):
        self.cnt += 1
        torch.save(x, self.file_base_name + str(self.cnt) + ".pt")

    def measure(self, x):
        self.update_state(x)
        self.used = True

    def is_used(self):
        return self.used


class ShapeObserver:
    def __init__(self, name, mod, d_shape=None, params=None):
        self.name = name
        self.mod = mod
        self.state = None

    def init_state(self, x):
        device = x.device
        Ndim = len(x.shape)
        self.Ndim = Ndim
        state = torch.tensor(x.shape, device=device, dtype=torch.int32).reshape((1, Ndim))
        return state

    def init_state_from_shape(self, x_shape, device="hpu"):
        logger.info("ShapeObserver doesn't support init_state_from_shape")
        return

    def update_state(self, x):
        logger.info("ShapeObserver doesn't support update_state")
        return

    def measure(self, x):
        self.state = self.init_state(x)

    def is_used(self):
        return self.state is not None


observer_types = {
    "shape": ShapeObserver,
    "maxabs": MaxAbsObserver,
    "maxabs_per_channel": MaxAbsPerChannelObserver,
    "save": SaveObserver,
}

observer_params = {
    "maxabs_per_channel": {
        "linear": ModuleConfig(({"dim": -1},), ({"dim": -1},), {"weight": {"dim": 0}}),
        "matmul": ModuleConfig(
            (
                {"dim": -1},
                {"dim": -2},
            ),
            ({"dim": -1},),
            None,
        ),
    }
}



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod

import torch.nn as nn

from .common import *


class QuantDequantBase(nn.Module):
    def __init__(self, lp_dtype, hp_dtype="", *args, **kwargs):
        super(QuantDequantBase, self).__init__(*args, **kwargs)
        self.lp_dtype = lp_dtype
        self.hp_dtype = hp_dtype

    @abstractmethod
    def forward(self, *args, **kwargs):
        pass

    def extra_repr(self) -> str:
        return f"lp_dtype={self.lp_dtype}, hp_dtype={self.hp_dtype}"


class QuantDequantNone(QuantDequantBase):
    def __init__(self, lp_dtype, hp_dtype, *args, **kwargs):
        super(QuantDequantNone, self).__init__(lp_dtype, hp_dtype, *args, **kwargs)

    def forward(self, *args, **kwargs):
        return args[0]

    def extra_repr(self) -> str:
        repr = super(QuantDequantNone, self).extra_repr()
        return f"{repr}, doesn't quantize nor dequantize"


class QuantInput(QuantDequantBase):
    def __init__(self, scale_inv, lp_dtype, hp_dtype, *args, **kwargs):
        super(QuantInput, self).__init__(lp_dtype, hp_dtype, *args, **kwargs)
        self.scale_inv = nn.Parameter(scale_inv)

    def forward(self, x):
        return cast_to_fp8_fcn(x, self.lp_dtype, self.scale_inv)

    def extra_repr(self) -> str:
        repr = super(QuantInput, self).extra_repr()
        return f"{repr}, scale_inv dtype={self.scale_inv.dtype}"


class DequantOutput(QuantDequantBase):
    def __init__(self, scale, lp_dtype, hp_dtype, *args, **kwargs):
        super(DequantOutput, self).__init__(lp_dtype, hp_dtype, *args, **kwargs)
        self.scale = nn.Parameter(scale)

    def forward(self, x):
        return cast_from_fp8_fcn(x, self.hp_dtype, self.scale)

    def extra_repr(self) -> str:
        repr = super(DequantOutput, self).extra_repr()
        return f"{repr}, scale dtype={self.scale.dtype}"



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor PyTorch load entry for all algorithms."""

import json
import os

from neural_compressor.common.utils import FP8_QUANT  # unified namespace
from neural_compressor.common.utils import load_config_mapping  # unified namespace
from neural_compressor.torch.quantization.config import (
    AutoRoundConfig,
    AWQConfig,
    FP8Config,
    GPTQConfig,
    HQQConfig,
    RTNConfig,
    TEQConfig,
)
from neural_compressor.torch.utils import LoadFormat

config_name_mapping = {
    FP8_QUANT: FP8Config,
}


def load(model_name_or_path, original_model=None, format="default", device="cpu", **kwargs):
    """Load quantized model.

    1. Load INC quantized model in local.
        case 1: WOQ
            from neural_compressor.torch.quantization import load
            load(model_name_or_path="saved_results", original_model=fp32_model)

        case 2: INT8/FP8
            from neural_compressor.torch.quantization import load
            load(model_name_or_path='saved_result', original_model=fp32_model)

        case 3: TorchScript (IPEX)
            from neural_compressor.torch.quantization import load
            load(model_name_or_path='saved_result')

    2. Load HuggingFace quantized model, including GPTQ models and upstreamed INC quantized models in HF model hub.
        case 1: WOQ
            from neural_compressor.torch.quantization import load
            load(model_name_or_path=model_name_or_path, format="huggingface")

    Args:
        model_name_or_path (str):  torch checkpoint directory or hugginface model_name_or_path.
            If 'format' is set to 'huggingface', it means the huggingface model_name_or_path.
            If 'format' is set to 'default', it means the 'checkpoint_dir'.
            Parameter should not be None. it coworks with 'original_model' parameter to load INC
            quantized model in local.
        original_model (torch.nn.module or TorchScript model with IPEX or fx graph with pt2e, optional):
            original model before quantization. Needed if 'format' is set to 'default' and not TorchScript model.
            Defaults to None.
        format (str, optional): 'defult' for loading INC quantized model.
            'huggingface' for loading huggingface WOQ causal language model. Defaults to "default".
        device (str, optional): 'cpu', 'hpu'. specify the device the model will be loaded to.
            currently only used for weight-only quantization.
        kwargs (remaining dictionary of keyword arguments, optional):
            remaining dictionary of keyword arguments for loading huggingface models.
            Will be passed to the huggingface model's `__init__` method, such as 'trust_remote_code', 'revision'.

    Returns:
        The quantized model
    """
    if format == LoadFormat.DEFAULT.value:
        from neural_compressor.common.base_config import ConfigRegistry

        qconfig_file_path = os.path.join(os.path.abspath(os.path.expanduser(model_name_or_path)), "qconfig.json")
        with open(qconfig_file_path, "r") as f:
            per_op_qconfig = json.load(f)

        if " " in per_op_qconfig.keys():  # ipex qconfig format: {' ': {'q_op_infos': {'0': {'op_type': ...
            from neural_compressor.torch.algorithms import static_quant

            return static_quant.load(model_name_or_path)
        elif "static_quant" in per_op_qconfig.keys() or "pt2e_dynamic_quant" in per_op_qconfig.keys():  # PT2E
            from neural_compressor.torch.algorithms import pt2e_quant

            return pt2e_quant.load(model_name_or_path)
        else:
            config_mapping = load_config_mapping(qconfig_file_path, ConfigRegistry.get_all_configs()["torch"])
            # select load function
            config_object = config_mapping[next(iter(config_mapping))]

            if isinstance(
                config_object, (RTNConfig, GPTQConfig, AWQConfig, TEQConfig, AutoRoundConfig, HQQConfig)
            ):  # WOQ
                from neural_compressor.torch.algorithms import weight_only

                qmodel = weight_only.load(model_name_or_path, original_model, format=LoadFormat.DEFAULT, device=device)
                return qmodel.to(device)

            original_model.qconfig = config_mapping
            if isinstance(config_object, FP8Config):
                # TODO: support loading FP8 model
                raise NotImplementedError("`load` function for FP8 model is not supported yet.")
    elif format == LoadFormat.HUGGINGFACE.value:
        # now only support load huggingface WOQ causal language model
        from neural_compressor.torch.algorithms import weight_only

        qmodel = weight_only.load(model_name_or_path, format=LoadFormat.HUGGINGFACE, device=device, **kwargs)
        return qmodel.to(device)
    else:
        raise ValueError("`format` in load function can only be 'huggingface' or 'default', but get {}".format(format))



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""AutoRound quantization."""
import copy
import json
import time
from typing import Union

import torch
from auto_round import AutoRound  # pylint: disable=E0401
from auto_round.export.export_to_itrex.export import pack_model  # pylint: disable=E0401

from neural_compressor.torch.algorithms import Quantizer
from neural_compressor.torch.utils import get_accelerator, logger

from .utility import CapturedDataloader, InputCaptureModule


class AutoRoundQuantizer(Quantizer):
    """AutoRound Quantizer."""

    def __init__(
        self,
        quant_config: dict = {},
        enable_full_range: bool = False,  ##for symmetric, TODO support later
        batch_size: int = 8,
        amp: bool = True,
        device: str = None,
        lr_scheduler=None,
        dataset: Union[str, list, tuple, torch.utils.data.DataLoader] = "NeelNanda/pile-10k",
        enable_quanted_input: bool = True,
        enable_minmax_tuning: bool = True,
        lr: float = None,
        minmax_lr: float = None,
        low_gpu_mem_usage: bool = False,
        iters: int = 200,
        seqlen: int = 2048,
        nsamples: int = 128,
        sampler: str = "rand",
        seed: int = 42,
        nblocks: int = 1,
        gradient_accumulate_steps: int = 1,
        not_use_best_mse: bool = False,
        dynamic_max_gap: int = -1,
        data_type: str = "int",
        scale_dtype: str = "fp16",
        quant_block_list: list = None,
        act_bits: int = 32,
        act_group_size: int = None,
        act_sym: bool = None,
        act_dynamic: bool = True,
        low_cpu_mem_usage: bool = False,
        **kwargs,
    ):
        """Init a AutQRoundQuantizer object.

        Args:
            quant_config (dict): Configuration for weight quantization (default is None).
            quant_config={
                        'layer1':##layer_name
                        {
                            'data_type': 'int',
                            'bits': 4,
                            'group_size': 32,
                            'sym': False,
                            'act_data_type': None,
                            'act_bits': 32,
                            'act_sym': None,
                            'act_dynamic': True,
                        }
                        ...,
                    }
                keys:
                    data_type (str): The data type to be used (default is "int").
                    bits (int): Number of bits for quantization (default is 4).
                    group_size (int): Size of the quantization group (default is 128).
                    sym (bool): Whether to use symmetric quantization. (default is None).
            bits (int): Number of bits for quantization (default is 4).
            group_size (int): Size of the quantization group (default is 128).
            sym (bool): Whether symmetric quantization is to be used (default is False).
            enable_full_range (bool): Whether to enable full range quantization (default is False).
            batch_size (int): Batch size for training (default is 8).
            amp (bool): Whether to use automatic mixed precision (default is True).
            device: The device to be used for tuning (default is "auto").
            lr_scheduler: The learning rate scheduler to be used.
            dataset (str): The default dataset name (default is "NeelNanda/pile-10k").
            enable_quanted_input (bool): Whether to use the output of the previous quantized block as
                                            the input for the current block (default is True).
            enable_minmax_tuning (bool): Whether to enable weight min-max tuning (default is True).
            lr (float): The learning rate (default is None, will be set to 1.0/iters).
            minmax_lr (float): The learning rate for min-max tuning
                                    (default is None, it will be set to lr automatically).
            low_gpu_mem_usage (bool): Whether to use low GPU memory (default is True).
            iters (int): Number of iterations (default is 200).
            seqlen (int): Data length of the sequence for tuning (default is 2048).
            nsamples (int): Number of samples (default is 128).
            sampler (str): The sampling method (default is "rand").
            seed (int): The random seed (default is 42).
            nblocks (int): Number of blocks (default is 1).
            gradient_accumulate_steps (int): Number of gradient accumulation steps (default is 1).
            not_use_best_mse (bool): Whether to use mean squared error (default is False).
            dynamic_max_gap (int): The dynamic maximum gap (default is -1).
            data_type (str): The data type to be used (default is "int").
            scale_dtype (str): The data type of quantization scale to be used (default is "float16"), different kernels
                have different choices.
            quant_block_list (list): A list whose elements are list of block's layer names to be quantized.
            act_bits (int): Number of bits for activation quantization. Default is 32.
            act_group_size (int): Group size for activation quantization. Default is None.
            act_sym (bool): Whether to use symmetric activation quantization. Default is None.
            act_dynamic (bool): Whether to use dynamic activation quantization. Default is True.

        Returns:
            The quantized model.
        """
        super().__init__(quant_config)
        self.tokenizer = None
        self.enable_full_range = enable_full_range
        self.batch_size = batch_size
        self.amp = amp
        self.device = get_accelerator(kwargs.pop("device", "auto")).current_device_name()
        self.lr_scheduler = lr_scheduler
        self.enable_quanted_input = enable_quanted_input
        self.enable_minmax_tuning = enable_minmax_tuning
        self.lr = lr
        self.minmax_lr = minmax_lr
        self.low_gpu_mem_usage = low_gpu_mem_usage
        self.iters = iters
        self.seqlen = seqlen
        self.nsamples = nsamples
        self.sampler = sampler
        self.seed = seed
        self.nblocks = nblocks
        self.gradient_accumulate_steps = gradient_accumulate_steps
        self.not_use_best_mse = not_use_best_mse
        self.dynamic_max_gap = dynamic_max_gap
        self.data_type = data_type
        self.scale_dtype = scale_dtype
        self.quant_block_list = quant_block_list
        self.act_bits = act_bits
        self.act_group_size = act_group_size
        self.act_sym = act_sym
        self.act_dynamic = act_dynamic
        self.low_cpu_mem_usage = low_cpu_mem_usage

    def prepare(self, model: torch.nn.Module, *args, **kwargs):
        """Prepares a given model for quantization.

        Args:
            model (torch.nn.Module): The model to be prepared.

        Returns:
            A prepared model.
        """
        prepare_model = InputCaptureModule(model)
        return prepare_model

    def convert(self, model: torch.nn.Module, *args, **kwargs):
        """Convert the prepared model to a quantized model.

        Args:
            model (torch.nn.Module): the prepared model

        Returns:
            The quantized model.
        """
        dataloader = CapturedDataloader(model.args_list, model.kwargs_list)
        model = model.orig_model
        rounder = AutoRound(
            model=model,
            tokenizer=None,
            dataset=dataloader,
            layer_config=self.quant_config or {},
            enable_full_range=self.enable_full_range,
            batch_size=self.batch_size,
            amp=self.amp,
            device=self.device,
            lr_scheduler=self.lr_scheduler,
            enable_quanted_input=self.enable_quanted_input,
            enable_minmax_tuning=self.enable_minmax_tuning,
            lr=self.lr,
            minmax_lr=self.minmax_lr,
            low_gpu_mem_usage=self.low_gpu_mem_usage,
            iters=self.iters,
            seqlen=self.seqlen,
            nsamples=self.nsamples,
            sampler=self.sampler,
            seed=self.seed,
            nblocks=self.nblocks,
            gradient_accumulate_steps=self.gradient_accumulate_steps,
            not_use_best_mse=self.not_use_best_mse,
            dynamic_max_gap=self.dynamic_max_gap,
            data_type=self.data_type,
            scale_dtype=self.scale_dtype,
            quant_block_list=self.quant_block_list,
            act_bits=self.act_bits,
            act_group_size=self.act_group_size,
            act_sym=self.act_sym,
            act_dynamic=self.act_dynamic,
            low_cpu_mem_usage=self.low_cpu_mem_usage,
        )
        model, weight_config = rounder.quantize()
        model.autoround_config = weight_config
        model = pack_model(model, weight_config, device=self.device, inplace=True)
        return model


def get_dataloader(tokenizer, seqlen, dataset_name="NeelNanda/pile-10k", seed=42, bs=8, nsamples=128):
    """Generate a DataLoader for calibration using specified parameters.

    Args:
        tokenizer (Tokenizer): The tokenizer to use for tokenization.
        seqlen (int): The exact sequence length. samples < seqlen will be dropped,
                      samples longer than seqlen will be truncated
        dataset_name (str, optional): The name of the dataset or datasets separated by commas.
                                     Defaults to "NeelNanda/pile-10k".
        split (str, optional): The data split to use. Defaults to None.
        seed (int, optional): The random seed for reproducibility. Defaults to 42.
        bs (int, optional): The batch size. Defaults to 4.
        nsamples (int, optional): The total number of samples to include. Defaults to 128.

    Returns:
        DataLoader: The DataLoader for the calibrated dataset.
    """
    from auto_round.calib_dataset import get_dataloader  # pylint: disable=E0401

    dataloader = get_dataloader(
        tokenizer, seqlen, dataset_name="NeelNanda/pile-10k", seed=seed, bs=bs, nsamples=nsamples
    )
    return dataloader



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from ..common import *
from ..fp_utils import *


def linear_unit_scale_scales(mod, measurement, params):
    device = torch.device("hpu")
    hp_dtype = params["hp_dtype"]
    input_scale = torch.tensor(1.0, dtype=hp_dtype, device=device)
    weight_scale = torch.tensor(1.0, dtype=hp_dtype, device=device)
    output_scale = torch.tensor(1.0, dtype=hp_dtype, device=device)
    return ModuleConfig((input_scale,), (output_scale,), {"weight": weight_scale})


def fsdpa_unit_scale_scales(mod, measurement, params):
    device = torch.device("hpu")
    hp_dtype = torch.float32  #  params["hp_dtype"]
    q_scale = torch.tensor(1.0, dtype=hp_dtype, device=device)
    k_scale = torch.tensor(1.0, dtype=hp_dtype, device=device)
    v_scale = torch.tensor(1.0, dtype=hp_dtype, device=device)
    softmax_scale = torch.tensor(1.0, dtype=hp_dtype, device=device)
    input_scale = (q_scale, k_scale, v_scale, softmax_scale)
    output_scale = (torch.tensor(1.0, dtype=hp_dtype, device=device),)
    return ModuleConfig(input_scale, output_scale, {})


def matmul_unit_scale_scales(mod, measurement, params):
    device = torch.device("hpu")
    hp_dtype = params["hp_dtype"]
    input_scale = (
        torch.tensor(1.0, dtype=hp_dtype, device=device),
        torch.tensor(1.0, dtype=hp_dtype, device=device),
    )
    output_scale = (torch.tensor(1.0, dtype=hp_dtype, device=device),)
    return ModuleConfig(input_scale, output_scale, {})


def softmax_unit_scale_scales(mod, measurement, params):
    device = torch.device("hpu")
    hp_dtype = params["hp_dtype"]
    input_scale = (torch.tensor(1.0, dtype=hp_dtype, device=device),)
    output_scale = (torch.tensor(1.0, dtype=hp_dtype, device=device),)
    return ModuleConfig(input_scale, output_scale)


def kv_cache_unit_scale_scales(mod, measurement, params):
    device = torch.device("hpu")
    hp_dtype = params["hp_dtype"]
    input_scale = (torch.tensor(1.0, dtype=hp_dtype, device=device),)
    output_scale = (torch.tensor(1.0, dtype=hp_dtype, device=device),)
    return ModuleConfig(input_scale, output_scale)



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import torch

from .._quant_common.quant_config import ScaleMethod, set_hqt_config
from ..utils.logger import logger
from .common import *
from .fp_utils import *
from .quant_dequant import *
from .scale_methods import *


def matmul_scales_to_mod_config(mod, scales, params):
    scales_inv = invert_scales(scales)
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    input_config = [QuantInput(s_inv, lp_dtype, hp_dtype) for s_inv in scales_inv.inputs]
    # outputs as bf16, and descaled in gemm under PatchedMatmul, so no need to work here
    output_config = [QuantDequantNone(lp_dtype, hp_dtype)]
    config = ModuleConfig(input_config, output_config, {})
    return config


def fsdpa_scales_to_mod_config(mod, scales, params):
    scales_inv = invert_scales(scales)
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    input_config = [QuantInput(s_inv, lp_dtype, hp_dtype) for s_inv in scales_inv.inputs]
    output_config = [DequantOutput(scales.outputs[0], lp_dtype, hp_dtype)]
    config = ModuleConfig(input_config, output_config, {})
    return config


def linear_scales_to_mod_config(mod, scales, params):
    scales_inv = invert_scales(scales)
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    input_config = [QuantInput(scales_inv.inputs[0], lp_dtype, hp_dtype)]
    # outputs as bf16, and descaled in gemm under PatchedLinear, so no need to work here
    output_config = [QuantDequantNone(lp_dtype, hp_dtype)]
    if isinstance(scales_inv.params["weight"], (torch.Tensor, float)):
        weight_config = QuantInput(scales_inv.params["weight"], lp_dtype, hp_dtype)
    elif isinstance(scales_inv.params["weight"], dict):
        weight_scale_inv_out_ch = scales_inv.params["weight"][0]
        weight_scale_inv_in_ch = scales_inv.params["weight"][1]
        if isinstance(weight_scale_inv_out_ch, torch.Tensor):
            scale_inv = torch.mul(
                weight_scale_inv_in_ch.reshape([1, -1]),
                weight_scale_inv_out_ch.reshape([-1, 1]),
            )
        else:
            # TODO SW-169781: Handle here scalar weight for PCQ
            raise TypeError(f"Unknown weight scales type: {type(weight_scale_inv_out_ch)}.")
        weight_config = QuantInput(scale_inv, lp_dtype, hp_dtype)
    else:
        logger.error("Unknown weight scales format.")
    params_config = {"weight": weight_config}
    if hasattr(mod, "bias") and (getattr(mod, "bias") is not None):
        # In PatchedLinear the bias is added to the output of gemm.
        # The output is expected to be descaled and in bf16, so we don't need to touch the bias.
        bias_config = QuantDequantNone(lp_dtype, hp_dtype)
        params_config.update({"bias": bias_config})
    config = ModuleConfig(input_config, output_config, params_config)
    return config


def kv_cache_scales_to_mod_config(mod, scales, params):
    # how quant/dequant will be applied on layer tensors
    scales_inv = invert_scales(scales)
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    input_config = [QuantInput(scales_inv.inputs[0], lp_dtype, hp_dtype)]
    output_config = [DequantOutput(scales.outputs[0], lp_dtype, hp_dtype)]
    config = ModuleConfig(input_config, output_config)
    return config


def softmax_scales_to_mod_config(mod, scales, params):
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    output_config = [DequantOutput(scales.outputs[0], lp_dtype, hp_dtype)]
    return ModuleConfig(None, output_config)


def get_config(
    model,
    measurement,
    mod_dict,
    method,
    params,
    scales_file=None,
    recalc_scales=False,
    mod_list=None,
):
    with torch.no_grad():
        top_level_config = get_hqt_config(model)
        qconfig = {UNMEASURED_MODELS: []}
        scales_file_format = np.ndarray  # file_functions[os.path.splitext(scales_file)[1]][0]
        scales_obj = (
            load_scales(scales_file + ".npz", scales_file_format)
            if (scales_file is not None) and not recalc_scales
            else {}
        )
        scales = convert_scales_to_tensors_dict(scales_obj, scales_file_format, params["hp_dtype"])
        model_dict = dict(model.named_modules())
        for mname in mod_list:
            mod = model_dict[mname]
            set_hqt_config(mod, top_level_config)  # set config in the module, as it consumed by the patched module
            mod_type_str = mod.__class__.__name__
            layer_type = mod_dict[mod_type_str].type
            if mname not in scales:
                logger.debug("Calculating scales for layer %s", mname)
                if mname not in measurement:
                    qconfig[UNMEASURED_MODELS].append(mname)
                    logger.debug(
                        "Layer '%s' has no measurements therefore it can't be quantized.",
                        mname,
                    )
                    continue
                layer_measure = measurement[mname]  # ModuleConfig() of measurements
                scales[mname] = method[layer_type][0](mod, layer_measure, params)  # ModuleConfig() of scales
                if scales_file is not None:
                    scales_obj[mname] = ModuleConfig(
                        **format_functions_rec((torch.Tensor, scales_file_format))(scales[mname].__dict__)
                    )

            logger.debug(
                "Preparing quantization functions for layer %s layer_type=%s",
                mname,
                layer_type,
            )
            mod_config = method[layer_type][1](mod, scales[mname], params)  # ModuleConfig() of QuantDequant
            mod_extra_config = ModuleExtraConfig(
                mod_config.inputs,
                mod_config.outputs,
                mod_config.params,
                scales[mname],
                params,
            )
            qconfig[mname] = mod_extra_config
        if scales_file is not None:
            save_scales(model, scales_obj, scales_file_format, scales_file + ".npz")
            save_scales(model, scales_obj, scales_file_format, scales_file + ".json")
    return qconfig


scaling_methods = {
    "unit_scale": {
        "linear": (linear_unit_scale_scales, linear_scales_to_mod_config),
        "matmul": (matmul_unit_scale_scales, matmul_scales_to_mod_config),
        "softmax": (softmax_unit_scale_scales, softmax_scales_to_mod_config),
        "kv_cache": (kv_cache_unit_scale_scales, kv_cache_scales_to_mod_config),
        "fused_sdpa": (fsdpa_unit_scale_scales, fsdpa_scales_to_mod_config),
    },
    "act_maxabs_pts_weight_maxabs_pts_pow2_hw": {
        "linear": (
            linear_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            linear_scales_to_mod_config,
        ),
        "matmul": (
            matmul_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            matmul_scales_to_mod_config,
        ),
        "kv_cache": (
            kv_cache_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            kv_cache_scales_to_mod_config,
        ),
        "softmax": (
            softmax_input_unit_output_maxabs_pts_hw_scales,
            softmax_scales_to_mod_config,
        ),
        "fused_sdpa": (
            fsdpa_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            fsdpa_scales_to_mod_config,
        ),
    },
    "act_maxabs_pts_weight_maxabs_pts_pow2": {
        "linear": (
            linear_act_maxabs_pts_weight_maxabs_pts_pow2_scales,
            linear_scales_to_mod_config,
        ),
        "matmul": (
            matmul_act_maxabs_pts_weight_maxabs_pts_pow2_scales,
            matmul_scales_to_mod_config,
        ),
    },
    "act_maxabs_pts_pow2_hw_weights_maxabs_pcs_pow2": {
        "linear": (
            linear_act_maxabs_pts_pow2_hw_weights_maxabs_pcs_pow2_scales,
            linear_scales_to_mod_config,
        ),
        "matmul": (
            matmul_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            matmul_scales_to_mod_config,
        ),
        # kv_cache is pts as op in hw doesn't work in pcs
        "kv_cache": (
            kv_cache_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            kv_cache_scales_to_mod_config,
        ),
        "fused_sdpa": (
            fsdpa_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            fsdpa_scales_to_mod_config,
        ),
    },
    "act_maxabs_pts_weight_opt_pts_pow2": {
        "linear": (
            linear_act_maxabs_pts_weight_opt_pts_pow2_scales,
            linear_scales_to_mod_config,
        ),
        "matmul": (
            matmul_act_maxabs_pts_weight_maxabs_pts_pow2_scales,
            matmul_scales_to_mod_config,
        ),
    },
    "act_maxabs_pts_weight_opt_pts_hw": {
        "linear": (
            linear_act_maxabs_pts_weight_opt_pts_hw_scales,
            linear_scales_to_mod_config,
        ),
        "matmul": (
            matmul_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            matmul_scales_to_mod_config,
        ),
        "softmax": (
            softmax_input_unit_output_maxabs_pts_hw_scales,
            softmax_scales_to_mod_config,
        ),
        "fused_sdpa": (
            fsdpa_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            fsdpa_scales_to_mod_config,
        ),
    },
    "act_maxabs_pts_pow2_hw_weights_opt_pcs_pow2": {
        "linear": (
            linear_act_maxabs_pts_pow2_hw_weights_opt_pcs_pow2_scales,
            linear_scales_to_mod_config,
        ),
        "matmul": (
            matmul_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            matmul_scales_to_mod_config,
        ),
        # kv_cache is pts as op in hw doesn't work in pcs
        "kv_cache": (
            kv_cache_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            kv_cache_scales_to_mod_config,
        ),
        "fused_sdpa": (
            fsdpa_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            fsdpa_scales_to_mod_config,
        ),
    },
    "act_maxabs_pts_pow2_weights_maxabs_pcs_pow2": {
        "linear": (
            linear_act_maxabs_pts_pow2_weights_maxabs_pcs_pow2_scales,
            linear_scales_to_mod_config,
        ),
        "matmul": (
            matmul_act_maxabs_pts_weight_maxabs_pts_pow2_scales,
            matmul_scales_to_mod_config,
        ),
        # kv_cache is pts as op in hw doesn't work in pcs
        "kv_cache": (
            kv_cache_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            kv_cache_scales_to_mod_config,
        ),
        "fused_sdpa": (
            fsdpa_act_maxabs_pts_weight_maxabs_pts_pow2_scales,
            fsdpa_scales_to_mod_config,
        ),
    },
    "act_maxabs_pts_pow2_weights_opt_pcs_pow2": {
        "linear": (
            linear_act_maxabs_pts_pow2_weights_opt_pcs_pow2_scales,
            linear_scales_to_mod_config,
        ),
        "matmul": (
            matmul_act_maxabs_pts_weight_maxabs_pts_pow2_scales,
            matmul_scales_to_mod_config,
        ),
        # kv_cache is pts as op in hw doesn't work in pcs
        "kv_cache": (
            kv_cache_act_maxabs_pts_pow2_weight_opt_pcs_pow2_scales,
            kv_cache_scales_to_mod_config,
        ),
        "fused_sdpa": (
            fsdpa_act_maxabs_pts_weight_maxabs_pts_pow2_scales,
            fsdpa_scales_to_mod_config,
        ),
    },
    "smoothquant_weights_opt_pow2": {
        "linear": (
            linear_smoothquant_weights_opt_pow2_scales,
            linear_scales_to_mod_config,
        ),
        "matmul": (
            matmul_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            matmul_scales_to_mod_config,
        ),
    },
    "smoothquant_weights_maxabs_pow2": {
        "linear": (
            linear_smoothquant_weights_maxabs_pow2_scales,
            linear_scales_to_mod_config,
        ),
        "matmul": (
            matmul_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            matmul_scales_to_mod_config,
        ),
    },
    "weaksmoothquant_weights_maxabs_pow2": {
        "linear": (
            linear_weaksmoothquant_weights_maxabs_pow2_scales,
            linear_scales_to_mod_config,
        ),
        "matmul": (
            matmul_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales,
            matmul_scales_to_mod_config,
        ),
    },
}

scale_method_mapping = {
    (ScaleMethod.UNIT_SCALE, "maxabs"): "unit_scale",
    (ScaleMethod.UNIT_SCALE, "maxabs_per_channel"): "unit_scale",
    (ScaleMethod.MAXABS_HW, "maxabs"): "act_maxabs_pts_weight_maxabs_pts_pow2_hw",
    (ScaleMethod.MAXABS_POW2, "maxabs"): "act_maxabs_pts_weight_maxabs_pts_pow2",
    (ScaleMethod.MAXABS_HW_OPT_WEIGHT, "maxabs"): "act_maxabs_pts_weight_opt_pts_hw",
    (
        ScaleMethod.MAXABS_POW2_OPT_WEIGHT,
        "maxabs",
    ): "act_maxabs_pts_weight_opt_pts_pow2",
    (
        ScaleMethod.ACT_MAXABS_HW_WEIGHTS_PCS_MAXABS_POW2,
        "maxabs",
    ): "act_maxabs_pts_pow2_hw_weights_maxabs_pcs_pow2",
    (
        ScaleMethod.ACT_MAXABS_HW_WEIGHTS_PCS_MAXABS_POW2,
        "maxabs_per_channel",
    ): "act_maxabs_pts_pow2_hw_weights_maxabs_pcs_pow2",
    (
        ScaleMethod.SMOOTHQUANT_WEIGHTS_OUTPUT_CHANNEL_MAXABS_POW2,
        "maxabs_per_channel",
    ): "smoothquant_weights_maxabs_pow2",
    (
        ScaleMethod.WEAKSMOOTHQUANT_WEIGHTS_OUTPUT_CHANNEL_MAXABS_POW2,
        "maxabs_per_channel",
    ): "weaksmoothquant_weights_maxabs_pow2",
    (
        ScaleMethod.ACT_MAXABS_HW_WEIGHTS_PCS_OPT_POW2,
        "maxabs",
    ): "act_maxabs_pts_pow2_hw_weights_opt_pcs_pow2",
    (
        ScaleMethod.ACT_MAXABS_HW_WEIGHTS_PCS_OPT_POW2,
        "maxabs_per_channel",
    ): "act_maxabs_pts_pow2_hw_weights_opt_pcs_pow2",
    (
        ScaleMethod.ACT_MAXABS_POW2_WEIGHTS_PCS_MAXABS_POW2,
        "maxabs",
    ): "act_maxabs_pts_pow2_weights_maxabs_pcs_pow2",
    (
        ScaleMethod.ACT_MAXABS_POW2_WEIGHTS_PCS_MAXABS_POW2,
        "maxabs_per_channel",
    ): "act_maxabs_pts_pow2_weights_maxabs_pcs_pow2",
    (
        ScaleMethod.ACT_MAXABS_POW2_WEIGHTS_PCS_OPT_POW2,
        "maxabs",
    ): "act_maxabs_pts_pow2_weights_opt_pcs_pow2",
    (ScaleMethod.SMOOTHQUANT_OPT, "maxabs_per_channel"): "smoothquant_weights_opt_pow2",
}

scaling_params = {
    "unit_scale": {},
    "act_maxabs_pts_weight_maxabs_pts_pow2_hw": {
        "input_backoff": 0.25,
        "weight_backoff": 0.5,
    },
    "act_maxabs_pts_weight_maxabs_pts_pow2": {
        "input_backoff": 0.25,
        "weight_backoff": 0.5,
    },
    "act_maxabs_pts_weight_opt_pts_pow2": {
        "input_backoff": 0.25,
        "weight_backoff": 0.5,
        "weight_scales": [2.0**s for s in range(-10, 10)],
    },
    "act_maxabs_pts_weight_opt_pts_hw": {
        "input_backoff": 0.25,
        "weight_backoff": 0.5,
        "weight_scales": [2.0**s for s in [4, 0, -4, -8]],
    },
    "smoothquant_weights_maxabs_pow2": {
        "input_backoff": 0.25,
        "weight_backoff": 0.5,
        "alpha": 0.5,
    },
    "weaksmoothquant_weights_maxabs_pow2": {
        "input_backoff": 0.25,
        "weight_backoff": 0.5,
        "alpha": 0.5,
    },
    "act_maxabs_pts_pow2_hw_weights_maxabs_pcs_pow2": {
        "input_backoff": 0.25,
        "weight_backoff": 0.5,
    },
    "act_maxabs_pts_pow2_hw_weights_opt_pcs_pow2": {
        "input_backoff": 0.25,
        "weight_backoff": 0.5,
        "weight_scales": [2.0**s for s in range(-3, 5)],
    },
    "act_maxabs_pts_pow2_weights_maxabs_pcs_pow2": {
        "input_backoff": 0.25,
        "weight_backoff": 0.5,
    },
    "act_maxabs_pts_pow2_weights_opt_pcs_pow2": {
        "input_backoff": 0.25,
        "weight_backoff": 0.5,
        "weight_scales": [2.0**s for s in range(-3, 5)],
    },
    "smoothquant_weights_opt_pow2": {
        "input_backoff": 0.25,
        "weight_backoff": 0.5,
        "alpha": 0.5,
        "transformed_weight_scales": [2.0**s for s in range(-3, 5)],
    },
}



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .._quant_common.helper_modules import *
from .._quant_common.quant_config import QuantMode, get_hqt_config
from ..utils.logger import logger
from .common import mod_default_dict
from .measure import prepare_model as prepare_model_for_measure
from .quantize import quantize
from .scale import scale_method_mapping, scaling_params


def update_mod_dict(config):
    assert (
        len(config.cfg["mod_dict"]) == 0
    ), f"Custom modules are not supported: {config.cfg['mod_dict'].keys()}. Please add it in the code."
    config.cfg["mod_dict"].update({k: mod_default_dict[k].type for k in mod_default_dict})


def print_init_info(config):
    import importlib.metadata

    versionStr = importlib.metadata.version("neural_compressor_pt")
    locationStr = versionStr.find("git") + 3
    logger.info("neural_compressor_pt Git revision = %s", versionStr[locationStr:])
    logger.info("neural_compressor_pt Configuration = %s", config)


def is_substr(substr_list, target):
    return any([x in target for x in substr_list])


def prepare_model(model):
    config = get_hqt_config(model)
    update_mod_dict(config)
    allowlist = set(config.cfg["mod_dict"].keys())
    blocklist = set()
    for type_st in config.cfg["blocklist"]["types"]:
        blocklist.add(type_st)
    allowlist.difference_update(blocklist)
    allowlist_tuple = tuple(allowlist)
    mod_list = []
    for name, mod in model.named_modules():
        mod_type = mod.__class__.__name__
        if (
            (mod_type in allowlist_tuple)
            and (is_substr(config.cfg["allowlist"]["names"], name) or len(config.cfg["allowlist"]["names"]) == 0)
            and (not is_substr(config.cfg["blocklist"]["names"], name))
        ):
            mod_list.append(name)

    print_init_info(config)

    logger.debug("Module list: %s", mod_list)
    logger.info("Total modules : %d", len(mod_list))
    if (config.cfg["mode"] == QuantMode.MEASURE) or (config.cfg["mode"] == QuantMode.SHAPE):
        return prepare_model_for_measure(model, mod_list)
    elif config.cfg["mode"] == QuantMode.QUANTIZE:
        scaling_method_name = scale_method_mapping[(config.cfg["scale_method"], config.cfg["observer"])]
        scaling_params[scaling_method_name].update(config.cfg["scale_params"])
        config.cfg["scale_params"] = scaling_params[scaling_method_name]
        return quantize(model, mod_list)



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

# For mapping revert patched module to origin module

helper_mods = {}


def helper_mod_register(name):
    def decorator(mod):
        helper_mods[name] = mod
        return mod

    return decorator


@helper_mod_register(name="Matmul")
class Matmul(torch.nn.Module):
    def __init__(self, patched_mod, *args, **kwargs):
        super().__init__()
        self.__dict__.update(patched_mod.__dict__)
        self.extra_repr = patched_mod.extra_repr_org


@helper_mod_register(name="Linear")
class Linear(torch.nn.Module):
    def __init__(self, patched_mod, *args, **kwargs):
        super().__init__()
        self.__dict__.update(patched_mod.__dict__)
        self.extra_repr = patched_mod.extra_repr_org


@helper_mod_register(name="FalconLinear")
class FalconLinear(torch.nn.Module):
    def __init__(self, patched_mod, *args, **kwargs):
        super().__init__()
        self.__dict__.update(patched_mod.__dict__)
        self.extra_repr = patched_mod.extra_repr_org


@helper_mod_register(name="KVCache")
class KVCache(torch.nn.Module):
    def __init__(self, patched_mod, *args, **kwargs):
        super().__init__()
        self.allocate = patched_mod.org_allocate
        self.get_shape = patched_mod.get_shape
        self.forward = patched_mod.forward
        self.update = patched_mod.update


@helper_mod_register(name="Conv2d")
class Conv2d(torch.nn.Module):
    def __init__(self, patched_mod, *args, **kwargs):
        super().__init__()
        self.__dict__.update(patched_mod.__dict__)
        self.extra_repr = patched_mod.extra_repr_org


@helper_mod_register(name="LoRACompatibleLinear")
class LoRACompatibleLinear(torch.nn.Module):
    def __init__(self, patched_mod, *args, **kwargs):
        super().__init__()
        self.__dict__.update(patched_mod.__dict__)
        self.extra_repr = patched_mod.extra_repr_org


@helper_mod_register(name="LoRACompatibleConv")
class LoRACompatibleConv(torch.nn.Module):
    def __init__(self, patched_mod, *args, **kwargs):
        super().__init__()
        self.__dict__.update(patched_mod.__dict__)
        self.extra_repr = patched_mod.extra_repr_org


@helper_mod_register(name="Softmax")
class Softmax(torch.nn.Module):
    def __init__(self, patched_mod, *args, **kwargs):
        super().__init__()
        self.__dict__.update(patched_mod.__dict__)
        self.extra_repr = patched_mod.extra_repr_org


@helper_mod_register(name="LinearLayer")
class LinearLayer(torch.nn.Module):
    def __init__(self, patched_mod, *args, **kwargs):
        super().__init__()
        self.__dict__.update(patched_mod.__dict__)
        self.extra_repr = patched_mod.extra_repr_org


@helper_mod_register(name="LinearAllreduce")
class LinearAllreduce(torch.nn.Module):
    def __init__(self, patched_mod, *args, **kwargs):
        super().__init__()
        self.__dict__.update(patched_mod.__dict__)
        self.extra_repr = patched_mod.extra_repr_org


@helper_mod_register(name="ScopedLinearAllReduce")
class ScopedLinearAllReduce(torch.nn.Module):
    def __init__(self, patched_mod, *args, **kwargs):
        super().__init__()
        self.__dict__.update(patched_mod.__dict__)
        self.extra_repr = patched_mod.extra_repr_org


@helper_mod_register(name="LmHeadLinearAllreduce")
class LmHeadLinearAllreduce(torch.nn.Module):
    def __init__(self, patched_mod, *args, **kwargs):
        super().__init__()
        self.__dict__.update(patched_mod.__dict__)
        self.extra_repr = patched_mod.extra_repr_org


@helper_mod_register(name="ModuleFusedSDPA")
class ModuleFusedSDPA(torch.nn.Module):
    def __init__(self, patched_mod, *args, **kwargs):
        super().__init__()
        self.__dict__.update(patched_mod.__dict__)
        self.extra_repr = patched_mod.extra_repr_org



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from tqdm import tqdm

from ..common import *
from ..fp_utils import *


def linear_smoothquant_weights_opt_pow2_scales(mod, measurement, params):
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    weight_backoff = params["weight_backoff"]
    alpha = params["alpha"]
    transformed_weight_scales = params["transformed_weight_scales"]
    input_range = torch.tensor(measurement.inputs[0], dtype=hp_dtype, device=device)
    weight_range_in_ch = torch.max(torch.abs(mod.weight), dim=0)[0].reshape([-1, 1])
    input_scale = calc_maxabs_scale(input_range, fullscale, input_backoff)
    weight_scale_in_ch = calc_maxabs_scale(weight_range_in_ch, fullscale, weight_backoff)
    input_scale = (input_scale**alpha) / (weight_scale_in_ch ** (1 - alpha))
    input_scale = scale_to_pow2(input_scale)
    weight_scale_in_ch = 1 / input_scale
    trans_weight = scale_fcn(mod.weight, weight_scale_in_ch.reshape([1, -1]))
    trans_weight_range_out_ch = torch.max(torch.abs(trans_weight), dim=1)[0].reshape([-1, 1])
    trans_weight_maxabs_scale_out_ch = calc_maxabs_scale(trans_weight_range_out_ch, fullscale, weight_backoff)
    trans_weight_maxabs_scale_out_ch = scale_to_pow2(trans_weight_maxabs_scale_out_ch)
    trans_weight_scale_out_ch = torch.zeros(mod.weight.shape[0])
    for k in tqdm(range(trans_weight_scale_out_ch.shape[0])):
        trans_weight_scale_out_ch[k] = mmse_scale(
            trans_weight[k, :],
            [s * trans_weight_maxabs_scale_out_ch[k] for s in transformed_weight_scales],
            lp_dtype,
            hp_dtype,
        )
    weight_scale_out_ch = scale_to_pow2(trans_weight_scale_out_ch)
    output_scale = torch.tensor(weight_scale_out_ch, dtype=hp_dtype, device=device)
    return ModuleConfig(
        (input_scale.flatten(),),
        (output_scale.flatten(),),
        {"weight": {0: weight_scale_out_ch.flatten(), 1: weight_scale_in_ch.flatten()}},
    )


def linear_smoothquant_weights_maxabs_pow2_scales(mod, measurement, params):
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    weight_backoff = params["weight_backoff"]
    alpha = params["alpha"]
    input_range = torch.tensor(measurement.inputs[0], dtype=hp_dtype, device=device)
    weight_range_in_ch = torch.max(torch.abs(mod.weight), dim=0)[0].reshape([-1, 1])
    input_scale = calc_maxabs_scale(input_range, 1.0, 1.0)
    weight_scale_in_ch = calc_maxabs_scale(weight_range_in_ch, 1.0, 1.0)
    input_scale = (input_scale**alpha) / (weight_scale_in_ch ** (1 - alpha))
    input_scale = scale_to_pow2(input_scale)
    input_range_post = input_range / input_scale
    input_scale_post = calc_maxabs_scale(input_range_post.max(), fullscale, input_backoff)
    input_scale_post = scale_to_pow2(input_scale_post)
    input_scale = input_scale * input_scale_post
    weight_scale_in_ch = 1 / input_scale
    trans_weight = scale_fcn(mod.weight, weight_scale_in_ch.reshape([1, -1]))
    trans_weight_range_out_ch = torch.max(torch.abs(trans_weight), dim=1)[0].reshape([-1, 1])
    trans_weight_maxabs_scale_out_ch = calc_maxabs_scale(trans_weight_range_out_ch, fullscale, weight_backoff)
    trans_weight_maxabs_scale_out_ch = scale_to_pow2(trans_weight_maxabs_scale_out_ch)
    weight_scale_out_ch = scale_to_pow2(trans_weight_maxabs_scale_out_ch)
    output_scale = torch.tensor(weight_scale_out_ch, dtype=hp_dtype, device=device)
    return ModuleConfig(
        (input_scale.flatten(),),
        (output_scale.flatten(),),
        {"weight": {0: weight_scale_out_ch.flatten(), 1: weight_scale_in_ch.flatten()}},
    )


def linear_weaksmoothquant_weights_maxabs_pow2_scales(mod, measurement, params):
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    weight_backoff = params["weight_backoff"]
    alpha = params["alpha"]
    input_range = torch.tensor(measurement.inputs[0], dtype=hp_dtype, device=device).max().clamp(min=1e-5)
    input_range_mid = input_range.max() / torch.sqrt(input_range.max() / input_range.min().clamp(min=1e-5))
    input_scale_pcs = calc_maxabs_scale(input_range.clamp(min=1e-5), input_range_mid, 1.0).clamp(min=1e-5)
    weight_range_in_ch = torch.max(torch.abs(mod.weight), dim=0)[0].reshape([-1, 1]).clamp(min=1e-5)
    weight_range_in_ch_mid = weight_range_in_ch.max() / torch.sqrt(
        weight_range_in_ch.max() / weight_range_in_ch.min().clamp(min=1e-5)
    ).clamp(min=1e-5)
    weight_scale_pcs = calc_maxabs_scale(weight_range_in_ch.clamp(min=1e-5), weight_range_in_ch_mid, 1.0).clamp(
        min=1e-5
    )

    input_scale = ((input_scale_pcs**alpha) / (weight_scale_pcs ** (1 - alpha))).clamp(min=1e-5)
    input_scale = scale_to_pow2(input_scale)
    input_scale_post = calc_maxabs_scale((input_range / input_scale).max(), fullscale, input_backoff)
    input_scale_post = scale_to_pow2(input_scale_post)

    weight_scale_in_ch = torch.ones([mod.weight.shape[1], 1], dtype=hp_dtype, device=device) * (1 / input_scale)

    trans_weight = scale_fcn(mod.weight, weight_scale_in_ch.reshape([1, -1]))
    weight_range_out_ch = torch.max(torch.abs(trans_weight), dim=1)[0].reshape([-1, 1])

    weight_maxabs_scale_out_ch = calc_maxabs_scale(weight_range_out_ch, fullscale, weight_backoff)
    weight_maxabs_scale_out_ch = scale_to_pow2(weight_maxabs_scale_out_ch)
    output_scale = weight_maxabs_scale_out_ch * input_scale_post
    return ModuleConfig(
        (input_scale.flatten() * input_scale_post,),
        (output_scale.flatten(),),
        {
            "weight": {
                0: weight_maxabs_scale_out_ch.flatten(),
                1: weight_scale_in_ch.flatten(),
            }
        },
    )
